module AE


  module ToolbarEditor


    module CustomButtons


      @@entries ||= {}


      # Gets all entries of custom button hashes.
      # @return [Array<Hash>]
      def self.get_all
        return @@entries.values.map{ |entry| entry.clone }
      end


      # Gets the entry hash of a specific custom button.
      # @param [String] id
      # @return [Hash]
      def self.get_by_id(id)
        return @@entries[id].clone if @@entries.include?(id)
      end


      # Loads a custom button from file.
      # @param [String] id
      # @param [Hash] metadata
      def self.load(id, metadata)
        # Load the code from a file.
        code = read_from_file(id)
        # Create the command.
        if code
          self.create(id, code, metadata)
        end
      end


      # Saves a new or modified custom button.
      # @param [String] id
      # @param [String] code
      # @param [Hash] metadata
      # @return [Boolean] success
      def self.save(id, code, metadata)
        if @@entries.include?(id)
          success = self.update(id, code, metadata)
        else
          success = self.create(id, code, metadata)
        end
        return success
      end


      # Creates a new custom button.
      # @param [String] id
      # @param [String] code
      # @param [Hash] metadata
      # @return [Boolean] success
      def self.create(id, code, metadata)
        return false unless id.is_a?(Fixnum) || id.is_a?(String)
        # Create the entry.
        entry = {:id => id}
        update_entry(entry, metadata)
        # Create the command.
        return false unless create_command(entry, code)
        # Write the code to a file.
        return false unless write_to_file(id, code, metadata)
        @@entries[id] = entry
        return true
      end


      # Updates an existing custom button.
      # @param [String] id
      # @param [String] code
      # @param [Hash] metadata
      # @return [Boolean] success
      def self.update(id, code, metadata)
        # Update the entry.
        entry = @@entries[id]
        return false unless entry.is_a?(Hash)
        update_entry(entry, metadata)
        # Update the command.
        return false unless update_command(entry, code)
        # Write the code to a file.
        return false unless write_to_file(id, code, metadata)
        return true
      end


      # Deletes a custom button.
      # @param [String] id
      def self.delete(id)
        @@entries.delete(id)
      end


      class << self


        # Updates the entry hash of a custom button.
        # @param [Hash] entry
        # @param [Hash] metadata
        def update_entry(entry, metadata)
          entry[:name] = metadata["name"]
          entry[:category] = metadata["category"] if metadata.include?("category")
          entry[:description] = metadata["description"] if metadata.include?("description")
          entry[:icon] = metadata["icon"] if metadata.include?("icon")
        end
        private :update_entry


        # Creates the command object of a custom button.
        # @param [Hash] entry
        # @param [String] code
        # @return [UI::Command,NilClass] depending on success
        def create_command(entry, code)
          name = entry[:name]
          raise ArgumentError.new("Argument 'code' must be a String and Hash 'entry' must contain 'name'") unless code.is_a?(String) && name.is_a?(String)
          # Create a Proc.
          begin
            block = eval(
              "Proc.new{
                begin
                  #{code}
                rescue Exception => e
                  e.message << ' in custom button #{name}'
                  if defined?(AE::ConsolePlugin)
                    AE::ConsolePlugin.error(e)
                  else
                    $stderr.write(e.message << $/)
                    $stderr.write(e.backtrace.join($/) << $/)
                  end
                end
              }", TOPLEVEL_BINDING
            )
          rescue SyntaxError => e
            e.message << " in custom button #{name}"
            if defined?(AE::ConsolePlugin)
              AE::ConsolePlugin.error(e)
            else
              $stderr.write(e.message << $/)
              $stderr.write(e.backtrace.join($/) << $/)
            end
            return nil
          end
          # Create a command.
          # We only bind a reference to the entry, and fetch everytime the current proc. 
          # This allows the command to run updated code.
          command = UI::Command.new(name){
            entry[:proc].call if entry.is_a?(Hash) && entry[:proc].is_a?(Proc)
          }
          command.tooltip = entry[:description] || ""
          command.status_bar_text = entry[:instruction] || ""
          command.large_icon = entry[:large_icon] || entry[:icon] || ""
          command.small_icon = entry[:small_icon] || entry[:icon] || ""
          entry[:proc] = block
          entry[:command] = command
          entry[:code] = code
          return command
        end
        # Since we can currently neither set nor read the Proc of a UI::Command, we 
        # use the same method to update a command by creating a new command instance.
        alias_method :update_command, :create_command
        private :create_command, :update_command


        # Reads the code snippet from a Ruby file.
        # @param [String] id
        # @return [String, NilClass] the code string
        def read_from_file(id)
          snippet_path = Dir.glob(File.join(DATA_PATH, "#{id}*.rb")).first
          return nil unless snippet_path && File.exists?(snippet_path)
          code = IO.read(snippet_path)
          # Drop comment lines.
          code = $' if code[/^=begin\n(?:.|\n\r?)*?\n=end\n/]
          return code
        end
        private :read_from_file


        # Write the code snippet to a file.
        # @param [String] id
        # @param [String] code
        # @return [Boolean] success whether the snippet was saved successfully
        def write_to_file(id, code, metadata)
          entry = @@entries[id] || metadata
          name = entry[:name] || metadata["name"] || ""
          Dir.mkdir(DATA_PATH) unless File.exists?(DATA_PATH)
          # Delete old file.
          Dir.glob(File.join(DATA_PATH, "#{id}*.rb")).each{ |old_file| 
            File.delete(old_file) rescue nil
          }
          # Write new file.
          path = File.join(DATA_PATH, "#{id}_#{name}.rb")
          File.open(path, "w") { |file|
            # Write some relevant metadata as comment into the file.
            file.puts("=begin")
            entry.each{ |key, value|
              next if !value.is_a?(String) || key == :code || key == "code"
              # If the value contains line breaks, indent every new line.
              value_lines = value.gsub(/\n\r?/, "\n" + " "*12)
              file.puts("#{key.to_s.rjust(12, ' ')}: #{value_lines}")
            }
            file.puts("=end")
            # Write the code to the file.
            file.write(code)
          }
          return true
        rescue Errno::ENOENT => e
          e.message << " Could not write #{path}"
          if defined?(AE::ConsolePlugin)
            AE::ConsolePlugin.error(e)
          else
            $stderr.write(e.message << $/)
            $stderr.write(e.backtrace.join($/) << $/)
          end
          return false
        end
        private :write_to_file


      end # class << self


    end


  end


end
