define(['bridge', 'translate', 'jquery', 'jqueryui', 'debounce', 'toolbar', 'toolbarbutton', 'custombutton', 'dialogbuttoneditor'], function (Bridge, Translate, $, _, debounce, Toolbar, ToolbarButton, CustomButton, dialogButtonEditor) {

  /**
   * @module
   */
  var self = {};


  // Objects to hold data.
  var Index = [];
  var availableButtons = [];
  var toolbars = [];
  var toolbarCounter = 0;


  /**
   * Escapes a String for generating a regular Expression.
   * @param {string} string
   * @returns {string}
   * @private
   */
  function escapeRegExp (string) {
    return string.replace(/[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, "\\$&");
  }

  function initializeUI () {
    // Allow to remove items from sortable.
    $("#buttons-available").droppable({
      accept: ".toolbar .toolbar-button",
      drop: function(event, ui) {
        ui.helper.remove();
        ui.draggable.remove();
      }
    });

    // Button to add a new toolbar
    $("#button-add-toolbar").on("click", function() {
      var toolbar = new Toolbar(toolbarCounter++)
      toolbar.activate();
      $(toolbar).on("delete", function () {
        toolbars.splice($.inArray(toolbar, toolbars), 1);
      });
      toolbars.push(toolbar);
    });

    // Search box // TODO: Does this sometimes not trigger in IE?
    $("#input-filter").on("keydown", debounce(function() {
        var searchWord = $("#input-filter").val();
        if (searchWord.length > 0) {
          // Hide all buttons.
          for (var i = 0; i < availableButtons.length; i++) {
            try { availableButtons[i].hide(); } catch (e) {}
          }
          var searchExp = new RegExp(escapeRegExp(searchWord), "i");
          // Show those buttons that match.
          $("#buttons-available li").each(function() {
            var match = ($(this).text().search(searchExp) !== -1);
            if (match) { $(this).show(); }
          });
        } else {
          // If no filter is applied, show all again.
          for (var i = 0; i < availableButtons.length; i++) {
            try { availableButtons[i].show(); } catch (e) {}
          }
        }
      })
    );
    $("#input-filter").on("click", function() {
        window.setTimeout(function () {
          var searchWord = $("#input-filter").val();
          if (searchWord.length == 0) {
            // If no filter is applied, show all again.
            for (var i = 0; i < availableButtons.length; i++) {
              try { availableButtons[i].show(); } catch (e) {}
            }
          }
        }, 0);
      }
    );

    // Button to add a new custom button
    dialogButtonEditor.initialize();
    $("#button-add-custom-button").on("click", function() { 
      dialogButtonEditor.show(function(entry) {
        var button = new CustomButton(entry);
        $(button).on("delete", function () {
          var index = 0;
          while (availableButtons[index] && availableButtons[index] != button) {
            index++;
          }
          availableButtons.splice(index, 1);
        });
        // Allow to drag a copy of the button into any toolbars.
        button.jQueryElement.draggable({
          connectToSortable: ".toolbar",
          // We need to put the helper outside of the overflow:scroll/hidden element.
          helper: "clone",
          appendTo: document.body,
          scroll: false,
          revert: "invalid",
          cursor: "move", cursorAt: { top: 10, left: 20 }
        });
        // Add the button to the list.
        availableButtons.push(button.jQueryElement.get(0));
        $("#buttons-available").prepend(button.jQueryElement.get(0));
      });
    });

    // Apply button
    $("#button-apply").on("click", function() {
      var data = {}, toolbar;
      for (var i = 0; i < toolbars.length; i++) {
        var toolbar = toolbars[i];
        if (toolbar.buttons.length > 0) {
          data[toolbar.id] = {"name": toolbar.name, "buttons": toolbar.buttons};
        };
      };
      Bridge.call("save_toolbars", data);
    });

    // Cancel button
    $("#button-cancel").on("click", function() {
      Bridge.call("close");
    });

    // Disable text selection on various UI elements.
    $("ul, li").disableSelection();
  }


  /**
   * Build a list of custom buttons.
   * @param {array<object>} array - list of objects <name,description,icon,id>
   */
  self.loadCustomButtons = function (array) {
    // Keep the button metadata in an index.
    Index = Index.concat(array);
    // Generate buttons.
    var buttons = [], button, entry;
    for (var i = 0; i < array.length; i++) {
      entry = array[i];
      var button = new CustomButton(entry);
      if (!button.jQueryElement) { continue; }
      // Allow to drag a copy of the button into any toolbars.
      button.jQueryElement.draggable({
        connectToSortable: ".toolbar",
        // We need to put the helper outside of the overflow:scroll/hidden element.
        helper: "clone",
        appendTo: document.body,
        scroll: false,
        revert: "invalid",
        cursor: "move", cursorAt: { top: 10, left: 20 }
      });
      buttons.push(button.jQueryElement);
    }
    // Add the buttons to the list.
    $("#buttons-available").prepend(buttons);
    availableButtons = availableButtons.concat(buttons);
  };


  /**
   * Build a list of available buttons.
   * @param {array<object>} array - list of objects <name,description,icon,id>
   */
  self.loadButtons = function(array) {
    if (!Object.prototype.toString.call(array) === '[object Array]') { return; }
    // Loading of buttons finished, now we can remove the indicator.
    $("#loading").remove();
    // Add a separator.
    array.unshift({
      id: 0,
      name: Translate.get("Separator"),
      description: Translate.get("Add a separator to group toolbar buttons."),
      icon: "../images/separator.png"
    });
    // Keep the button metadata in an index.
    Index = Index.concat(array);
    // Generate buttons.
    var boo = true;
    // Add all buttons asynchronously in chunks of 20.
    for (var i = 0; i < array.length; i += 20) {
      (function (i) {
        setTimeout(function(){
          // Add a chunk of 20 buttons.
          var buttons = [];
          for (var j = i; j < i+20 && j < array.length; j++) {
            // Add one button.
            var entry = array[j];
            if (!entry) { continue; }
            var button = new ToolbarButton(entry.id, entry);
            if (!button.jQueryElement) { return; }
            // Allow to drag a copy of the button into any toolbars.
            button.jQueryElement.draggable({
              connectToSortable: ".toolbar",
              // We need to put the helper outside of the overflow:scroll/hidden element.
              helper: "clone",
              appendTo: document.body,
              scroll: false,
              revert: "invalid",
              cursor: "move", cursorAt: { top: 10, left: 20 }
            });
            // Add the button to the list.
            buttons.push(button.jQueryElement);
          }
          $("#buttons-available").append(buttons);
          availableButtons = availableButtons.concat(buttons);
        }, 0);
      })(i);
    }
  };


  /**
   * Build a list of created toolbars.
   * @param {array<object>} array - list of objects <name,description,icon,id>
   */
  self.loadToolbars = function(array) {
    if (Object.prototype.toString.call(array) !== '[object Array]') { return; }
    var entry, toolbar, buttons;
    for (var i = 0; i < array.length; i++) {
      entry = array[i];
      if (entry.id >= toolbarCounter) { toolbarCounter = entry.id + 1; }
      toolbar = new Toolbar(entry.id, entry.name);
      $(toolbar).on("delete", function () {
        toolbars.splice($.inArray(toolbar, toolbars), 1);
      });
      toolbars.push(toolbar);
      buttons = entry.buttons;
      for (var j = 0; j < buttons.length; j++) {
        var id = buttons[j];
        var entry = $.grep(Index, function (item){
          return item.id == id;
        })[0];
        if (!entry) continue;
        toolbar.addButton(id, entry);
      }
    }
    // Create a new toolbar if none exists.
    if (array.length == 0) {
      toolbar = new Toolbar(toolbarCounter++);
      $(toolbar).on("delete", function () {
        toolbars.splice($.inArray(toolbar, toolbars), 1);
      });
      toolbars.push(toolbar);
    }
    // Select the last toolbar.
    if (toolbars.length > 0) {
      toolbars[toolbars.length - 1].activate();
    }
  };


  /**
   * jQuery on document ready
   */
  $(initializeUI);


  return self;
});
