=begin
Usage:        Menu 'Window' → 'ToolbarEditor'
              Opens a dialog where you can drag and drop toolbar buttons from the
              list on the right into toolbar panels on the left. Click "Apply" to
              save the changes. You need to restart SketchUp for all changes to
              take effect.

Description:  This plugin lets you create your own toolbars.

Required:     SketchUp 8 M1 or higher

Information:  The principle of custom toolbars is to save&read the defined toolbars
              from a file, then loop over each toolbar configuration and create
              a new UI::Toolbar and add buttons for its commands.
              However there is no standard way in SketchUp to detect features/
              commands of other plugins and execute them (UI::Command.proc.call).
              One way is to write code for each plugin's method name manually in
              a configuration file or script (so the user would need to look 
              into each plugin's source code).
              Another way is to get all UI::Command instances from Ruby's ObjectSpace.
              Unfortunately UI::Command don't have persistent identifiers to
              remember them over SketchUp sessions. The Plugin LaunchUp therefore
              creates an index and generates a unique and persistent hash from
              each command's metadata. This plugin uses LaunchUp's index (if installed)
              or a lightweight substitute.
              Because of that it is important that the interception (in LaunchUp)
              happens before other scripts load, and that custom toolbars are
              created after all other scripts have loaded.
=end
require 'sketchup.rb'


module AE


  module ToolbarEditor

    # This plugin's folder.
    PATH = File.dirname(__FILE__) unless defined?(self::PATH)

    # Requirements
    %w(translate.rb
       bridge.rb
       options.rb
       ui.rb
       toolbars.rb
       custombuttons.rb
    ).each{ |file| require(File.join(PATH, file)) }

    EDITOR_HTML = File.join(PATH, "html", "ToolbarEditor.html") unless defined?(self::EDITOR_HTML)

    # Load translation strings and add plugin to UI.
    self::TRANSLATE = Translate.new('toolbareditor.strings') unless defined?(self::TRANSLATE)

    # Index of all commands.
    require(File.join(PATH, 'index.rb'))

    # Options.
    @options ||= Options.new(ToolbarEditor, {
      :toolbars => {},       # id => { "name" => [String], "buttons" => [Array<Fixnum>] }
      :custom_buttons => {}, # id => { "name" => [String], … }
    }).migrate("0.0.0", "1.1.1"){ |key, value|
      case key
      when :toolbars then
        new_value = {}
        value.each{ |hash|
          id = (hash["id"].is_a?(Fixnum)) ? hash["id"] : (!new_value.keys.empty?) ? new_value.keys.max.next : 0
          hash["id"] = id
          new_value[id] = hash
        }
        [key, new_value]
      when :snippets then
        new_value = {}
        value.each{ |hash|
          id = (hash["id"].is_a?(Fixnum)) ? hash["id"] : (!new_value.keys.empty?) ? new_value.keys.max.next : 0
          hash["id"] = id
          new_value[id] = hash
        }
        [:custom_buttons, new_value]
      else
        [key, value]
      end
    }

    # Reference to the main dialog.
    @editor_dlg ||= nil

    # Data directory for saving files.
    dir = [
           File.join(PATH, "data"),                         # Plugins folder
           ENV["APPDATA"],                                 # Windows
           File.join(ENV["HOME"].to_s, ".local", "share"), # Free desktop standard
           File.join(ENV["HOME"].to_s, "Library", "Application Support"), # OS X
           File.expand_path(".")                           # Fallback: user's folder.
          ].compact.find{ |path| File.exists?(path) && File.writable?(path) }
    DATA_PATH = File.join(File.expand_path(dir), "SketchUp ToolbarEditor") unless defined?(self::DATA_PATH)


    public


    # This displays the main dialog with search field.
    def self.show_dialog
      # If the dialog exists, bring it to the front.
      if @editor_dlg && @editor_dlg.visible?
        @editor_dlg.bring_to_front
      else

        properties = {
            :dialog_title    => TRANSLATE["Toolbar Editor"],
            :preferences_key => "com.aerilius.toolbareditor",
            :scrollable      => false,
            :resizable       => true,
            :width           => 600,
            :height          => 550,
            :left            => 600,
            :top             => 200,
        }
        if defined?(UI::HtmlDialog)
          properties[:style] = UI::HtmlDialog::STYLE_DIALOG
          @editor_dlg = UI::HtmlDialog.new(properties)
        else
          @editor_dlg = UI::WebDialog.new(properties)
        end
        @editor_dlg.set_file(EDITOR_HTML)

        # Add a Bridge to handle JavaScript-Ruby communication.
        @editor_dlg = Bridge.decorate(@editor_dlg)

        @editor_dlg.on("translate") { |action_context|
          # Translate.
          TRANSLATE.webdialog(@editor_dlg)
        }

        @editor_dlg.on("get_custom_buttons") { |action_context|
          action_context.resolve(CustomButtons.get_all)
        }

        @editor_dlg.on("get_available_buttons") { |action_context|
          buttons = Index.instance.get_all{ |hash|
            hash[:icon].is_a?(String) && !hash[:icon].empty? &&
            # If the LaunchUp Index is used, it will also include duplicates of the custom buttons.
            !CustomButtons.get_all.any?{ |e| e[:command] == hash[:command] }
          }.sort_by{ |entry|
            entry[:name] || ""
          }.sort_by{ |entry|
            entry[:category] || ""
          }
          action_context.resolve(buttons)
        }

        @editor_dlg.on("get_toolbars") { |action_context|
          action_context.resolve(Toolbars.get_all)
        }

        # Create a new custom button.
        @editor_dlg.on("save_custom_button") { |action_context, id, code, metadata|
          success = CustomButtons.save(id, code, metadata)
          if success
            # Save the data.
            @options[:custom_buttons][id] = metadata
            @options.save
          end
          action_context.resolve(success)
        }

        # Delete a new custom button.
        @editor_dlg.on("delete_custom_button") { |action_context, id|
          CustomButtons.delete(id)
          # Save the data.
          @options[:custom_buttons].delete(id)
          @options.save
        }

        # Save the toolbars and close the dialog.
        @editor_dlg.on("save_toolbars") { |action_context, hash|
          hash2 = {}
          hash.each{ |id, metadata|
            id = id.to_i
            Toolbars.save(id, metadata)
            hash2[id] = metadata
          }
          # Save the data.
          @options[:toolbars] = hash2
          @options.save
          @editor_dlg.close
        }

        # Delete a toolbar.
        @editor_dlg.on("delete_toolbar") { |action_context, id|
          Toolbars.delete(id)
          # Save the data.
          @options[:toolbars].delete(id)
          @options.save
        }

        # File dialog.
        @editor_dlg.on("openpanel") { |action_context, path|
          new_path = UI.openpanel(path)
          new_path = File.expand_path(new_path) if new_path
          action_context.resolve(new_path)
        }

        # Close the dialog.
        @editor_dlg.on("close") { |action_context|
          @editor_dlg.close
        }

        # Show the webdialog.
        @editor_dlg.show

      end
      return @editor_dlg
    end


    # Closes the main dialog if it is visible.
    def self.close
      if @editor_dlg && @editor_dlg.visible?
        @editor_dlg.close
      end
    end


    # Creates toolbar objects for commands remembered from @options and identified by IDs.
    def self.load
      if defined?(AE::LaunchUp) && defined?(AE::LaunchUp::Index) &&
          AE::LaunchUp::Index.method_defined?(:get_all)
        AE::LaunchUp::Index.instance.get_all.each{ |metadata|
          AE::ToolbarEditor::Index.instance.add(metadata)
        }
      end
      if !@options[:migrated_hashcodes_adler32_to_md5]
        self.migrate_hashcodes_adler32_to_md5(@options)
        @options[:migrated_hashcodes_adler32_to_md5] = true
        @options.save
      end

      # Load custom buttons.
      @options[:custom_buttons].keys.to_a.each{ |id|
        metadata = @options[:custom_buttons][id]
        metadata[:id] = metadata[:id].to_s
        if @options[:custom_buttons].delete(id)
          @options[:custom_buttons][id.to_s] = metadata
        end
      }
      @options[:custom_buttons].each{ |id, metadata|
        next unless metadata.is_a?(Hash) && metadata["name"].is_a?(String)
        @options[:custom_buttons][id] = metadata
        CustomButtons.load(id, metadata)
      }
      # Load toolbars.
      @options[:toolbars].each{ |id, metadata|
        next unless metadata.is_a?(Hash) && metadata["name"].is_a?(String) &&
          metadata["buttons"].is_a?(Array) && !metadata["buttons"].empty?
        Toolbars.load(id, metadata)
      }
    end


    def self.migrate_hashcodes_adler32_to_md5(options)
      options[:toolbars].each{ |toolbar_id, toolbar_dict|
        toolbar_dict["buttons"].map{ |old_button_id|
          new_button_id = Index.instance.convert_old_hash_code(old_button_id)
          new_button_id
        }
      }
    end


    unless file_loaded?(__FILE__)
      initialize_ui()

      # Load all custom toolbars when SketchUp starts, but only after all other plugins have been loaded.
      # Alternative 1: require all in load path:
=begin
      $LOAD_PATH.each{ |path|
        next unless path.is_a?(String) && File.directory?(path)
        # The following freezes SketchUp:
        Dir[File.join(path, "*.{rb,rbs}")][0,10].each{ |file|
          next if file == __FILE__ || $".include?(file)
          begin
            require(file)
          rescue
            next
          end
        }
      }
      self.load()
=end
      # Alternative 2: use timer that will execute after loading has finished.
      # Problem: SketchUp2013 doesn't remember UI::Toolbar.get_last_state for toolbars
      # that were created long after loading finished. This was fixed in 2013.
      UI.start_timer(0, false) {
        begin
          self.load()
        rescue Exception => error
          if defined?(AE::ConsolePlugin)
            AE::ConsolePlugin.error(error)
          else
            $stderr.write(error.message + $/)
            $stderr.write(error.backtrace.join($/) + $/)
          end
        end
      }

      file_loaded(__FILE__)
    end # end unless


  end


end
