module AE


  module ToolbarEditor


    module Toolbars


      @@entries ||= {}
      @@toolbars ||= {}


      # Gets all entries of toolbars hashes.
      # @return [Array<Hash>]
      def self.get_all
        return @@entries.values.map{ |entry| entry.clone }
      end


      # Saves a new or modified toolbar.
      # @param [Fixnum] id
      # @param [Hash] metadata
      def self.save(id, metadata)
        if @@entries.include?(id) && @@toolbars.include?(id)
          self.update(id, metadata)
        else
          self.create(id, metadata)
        end
      end


      # We can use the same method when loading toolbars on startup
      def self.load(id, metadata)
        self.save(id, metadata)
      end


      # Creates a new toolbar.
      # @param [Fixnum] id
      # @param [Hash] metadata
      def self.create(id, metadata)
        # Create the entry.
        entry = {:id => id}
        update_entry(entry, metadata)
        # Create the toolbar.
        toolbar = create_toolbar(metadata["name"], metadata["buttons"])
        # toolbar.show
        toolbar.restore
        # Per bug 2902434, adding a timer call to restore the toolbar. This
        # fixes a toolbar resizing regression on Windows as the restore() call
        # does not seem to work as the script is first loading.
        UI.start_timer(0.1, false) {
          toolbar.restore
        }
        toolbar.show
        @@toolbars[id] = toolbar
        @@entries[id] = entry
      end


      # Updates an existing toolbar.
      # @param [Fixnum] id
      # @param [Hash] metadata
      def self.update(id, metadata)
        # Update the entry.
        update_entry(@@entries[id], metadata)
        # Update the toolbar.
        update_toolbar(id, metadata["name"], metadata["buttons"])
      end


      # Deletes a toolbar.
      # @param [Fixnum] id
      def self.delete(id)
        @@entries.delete(id)
        @@toolbars.delete(id)
      end


      class << self


        # Updates the entry hash of a toolbar.
        # @param [Hash] entry
        # @param [Hash] metadata
        def update_entry(entry, metadata)
          entry.merge!(metadata)
        end
        private :update_entry


        # Creates a toolbar UI object.
        # @param [String] name
        # @param [Array<Fixnum>] button_ids
        # @return [UI::Toolbar]
        def create_toolbar(name, button_ids)
          # Create a new toolbar and add the commands.
          toolbar = UI::Toolbar.new(name)
          button_ids.each{ |id|
            if id == 0
              # Add a separator
              toolbar.add_separator()
            else
              # Get the entry.
              entry = Index.instance.get_by_id(id)
              entry = CustomButtons.get_by_id(id) if entry.nil?
              next if entry.nil?
              command = entry[:command] || UI::Command.new(entry[:name], &entry[:proc]) rescue next
              toolbar.add_item(command)
            end
          }
          return toolbar
        end
        private :create_toolbar


        # Updates the toolbar UI object.
        # Since we can currently not remove commands from toolbars, we only add new ones.
        # @param [Fixnum] id
        # @param [String] name
        # @param [Array<Fixnum>] button_ids
        def update_toolbar(id, name, button_ids)
          toolbar = @@toolbars[id]
          # toolbar.name = name
          # Get all commands that this toolbar already has, ignore strings.
          # Note: Iterating a toolbar returns new, different command references, and a 
          # direct comparison to references to another instance of the "same" command fails.
          commands = toolbar.entries
          # Check if there are new buttons.
          # if commands.length < button_ids.length
          button_ids.each{ |id|
            if id == 0
              # Add a separator
              toolbar.add_separator() unless commands.include?("|")
              next
            else
              # Get the entry.
              entry = Index.instance.get_by_id(id)
              entry = CustomButtons.get_by_id(id) if entry.nil?
              next if entry.nil?
              # Don't add it if it already is on the toolbar. 
              # Direct comparison fails, so we compare the name.
              next if commands.find{ |c| c == entry[:command] || c.is_a?(UI::Command) && c.menu_text == entry[:name] }
              # Add the command.
              command = entry[:command] || UI::Command.new(entry[:name], &entry[:proc]) rescue next
              toolbar.add_item(command)
            end
          }
          # end
        end
        private :update_toolbar


      end # class << self


    end


  end


end
